@@ -668,7 +668,7 @@ gnome_rr_screen_new (GdkScreen *gdk_screen,
 	
 	screen->gdk_screen = gdk_screen;
 	screen->gdk_root = gdk_screen_get_root_window (gdk_screen);
-	screen->xroot = gdk_x11_drawable_get_xid (screen->gdk_root);
+	screen->xroot = gdk_x11_window_get_xid (screen->gdk_root);
 	screen->xdisplay = dpy;
 	screen->xscreen = gdk_x11_screen_get_xscreen (screen->gdk_screen);
 	screen->connector_type_atom = XInternAtom (dpy, "ConnectorType", FALSE);
